/* SACHICA: Scalable Algorithm for Characteristic and Homologinous Interval CAlclation */
/* 2006 Takeaki Uno */
/* matdraw */

#ifndef _sachicam_c_
#define _sachicam_c_

#include"sachica.h"
#include"pos.c"
#include"problem.c"
  
/* show error messages */
void SACHICAm_error (){
  ERROR_MES = "command explanation";
  print_err (" SACHICAM: Scalable Algorithm for Charactaristic and Homologineous Interval CAlculation: draw a matrix or line of similarity, Ver1.1\n\
sachicam [options] input-file x-size y-size output-file\n\
-N: negative imagem, -t [threshold]: monotone with [threshold]\n\
-Z [num]: emphasize each dot by replacing to a box of size [num]\n\
-c [num]: comparison between 1st to [num]-1th, and the rest\n\
-k [real-num]: gray scale with brightness [real-num]\n\
-M: superimpose the diagonal flipped image\n\
-p: output bmpfile\n\
-H [position file]: read [position file]\n\
-h [position file]: draw grid lines according to [position file]\n\
-r [filename]: overlap x-coordinate reversed string in [filename]\n\
-R [filename]: overlap y-coordinate reversed string in [filename]\n\
-T [threshold]: threshold for length of x&y directions (x-x or y-yy >= th)\n\
-L [column num] [num]: line mode with line interval [num] (numbers in [column num]-th column will be dealt, if -1 is given, consider both 1st and 2nd columns)\n\
-l [num]: line mode with inputting interval format, with [num]\n\
-A: read the file as position-score file (has to be with -l)\n\
-C [len] [mis] [filename]: detect intervals s.t. each segment of length [len] has no less than [len]-[mis] dots, and write them to [filename]\n\
-d [filename]: detect pseudo diagonal lines and write to [filename]\n\
-e [filename]: detect pseudo diagonal lines and write to [filename] (reverse direction of x)\n\
-E [filename]: detect pseudo diagonal lines and write to [filename] (reverse direction of y)\n\
-S:long skip mode for pseudo diagonal line detection\n\
-b,-B [ratio]: (with -d,-e or, -E) output homologinous strings whose one(two if -B) end is closed to boundary with [ratio]\n\
if y-size is 0, y-size is determined by x-size and ratio of the real x,y lengths\n\
if y-size is -1, x-size specifies the size of a cell\n\
# the 1st letter of input-filename cannot be '-'.\n");
  EXIT;
}

/* read parameters from command line */
void SACHICAm_read_param (SACHICA *SA, int argc, char *argv[]){
  int c=1;
  SEQ *SS = &SA->PP.SS;
  PROBLEM *PP = &SA->PP;

  SA->th = 0;    // threshold for intensity of a pixel
  SA->th1 = 0;   // threshold for #touching boundary
  SA->th2 = 0;   // threshold for length of homology string
  SS->err_ratio = 1.0;  // ambiguity ratio for touching boundary 
  SA->density = 1.0;     // intensity
  SA->col_num = 0;     // column number to be written for line mode
  SA->line_skip = 0;
  SA->emphasize = 0;
  
  if ( argc<c+4 ){ SACHICAm_error(); return; }

  while ( argv[c][0] == '-' ){
    switch (argv[c][1]){
      case 't': SA->th = atoi(argv[c+1]);
      break; case 'T': SA->th2 = atoi(argv[c+1]);
      break; case 'S': PP->problem |= SACHICAM_LONG_SKIP; c--;
      break; case 'N': PP->problem |= SACHICAM_NEGATIVE; c--;
      break; case 'A': PP->problem |= SACHICAM_POSITION_SC; SA->col_num = 1; c--;
      break; case 'L': SA->col_num = atoi(argv[c+1]);
        if ( (SA->line_skip = atoi(argv[c+2])) < 0 )
            error_num("line interval length has to be positive",atoi(argv[c+2]), EXIT);
        c++;
      break; case 'l': if ( (SA->line_skip = atoi(argv[c+1])) < 0 )
            error_num("line interval length has to be positive",atoi(argv[c+1]), EXIT);
        PP->problem |= SACHICAM_READ_INTERVAL;
      break; case 'c': if ( (SS->separated_by = atoi(argv[c+1])) <= 0 )
          error_num("string ID has to be positive", atoi(argv[c+1]), EXIT);
      break; case 'Z': if ( (SA->emphasize = atoi(argv[c+1])) < 0 )
          error_num("empahsis intensity has to be positive",atoi(argv[c+2]), EXIT);
      break; case 'M': PP->problem |= SACHICAM_MIRROR; c--;
      break; case 'p': PP->problem |= SACHICAM_BMP; c--;
      break; case 'k': if ( (SA->density = atof(argv[c+1])) <= 0 )
          error_num("brightness has to be positive",atoi(argv[c+1]), EXIT);
      break; case 'H': PP->position_fname = argv[c+1];
      break; case 'h': PP->position_fname = argv[c+1]; PP->problem |= SACHICAM_DRAW_GRID;
      break; case 'r': PP->problem |= SACHICA_REVERSE_X; PP->position2_fname = argv[c+1];
      break; case 'R': PP->problem |= SACHICA_REVERSE_Y; PP->position2_fname = argv[c+1];
      break; case 'd': PP->problem |= SACHICAM_DETECT; PP->sc_fname = argv[c+1];
      break; case 'e': PP->problem |= SACHICAM_HOMOLOGY_REVERSE_X; PP->sc_fname = argv[c+1];
      break; case 'E': PP->problem |= SACHICAM_HOMOLOGY_REVERSE_Y; PP->sc_fname = argv[c+1];
      break; case 'b': SS->err_ratio = atof (argv[c+1]); SA->th1 = 1;
      break; case 'B': SS->err_ratio = atof (argv[c+1]); SA->th1 = 2;
      break; case 'C': SS->plen = atoi(argv[c+1]); SS->perr = atoi(argv[c+2]);
          PP->sc_fname = argv[c+3]; c+=2;
      break; default: goto NEXT;
    }
    c += 2;
    if ( argc < c+4 ){ SACHICAm_error(); return; }
  }

  NEXT:;
  PP->input_fname = argv[c];
  if ( (SA->pxmax = atoi(argv[c+1])) <= 0 )
        error_num("#horizontal pixels has to be positive",atoi(argv[c+2]), EXIT);
  if ( (SA->pymax = atoi(argv[c+2])) < -1 )
        error_num("#virtical pixels has to be positive, -1 or 0",atoi(argv[c+2]), EXIT);
  PP->output_fname = argv[c+3];
  
  // error for illegal parameters
  if ( (!PP->position_fname) && PP->position2_fname )
      error ("position file is needed for r,R options", EXIT);
}


/*****************************************/
void SACHICAm_scan_file (SACHICA *SA, char *fname){
  SEQ_LONG x, y, z, mm, zz, m=0, c=2;
  FILE2 fp;
  PROBLEM *PP = &SA->PP;
  SEQ *SS = &PP->SS;
  SA->xmax = SA->ymax = 0;
  
  if ( PP->position_fname ){
    FILE2_open (fp, PP->position_fname, "r", EXIT);
    z = (SEQ_LONG)FILE2_read_int (&fp);
    mm = (SEQ_LONG)FILE2_read_int (&fp);
    FILE2_read_until_newline (&fp);
    while (1){
      z = (SEQ_LONG)FILE2_read_int (&fp);
      zz = (SEQ_LONG)FILE2_read_int (&fp);
      FILE2_read_until_newline (&fp);
      if ( FILE_err&2 ) break;
      if ( c == SS->separated_by ) mm = z;
      ENMAX (m, zz);
      c++;
    }
    if ( m == 0 ) m = mm;
    if ( c>2 && (SA->pymax==0 || SS->separated_by>0) ){  // comparing two files
      SA->xmax = mm; SA->ymax = m-mm;
    } else {
      SA->xmax = SA->ymax = m;
      if ( SA->pymax == 0 ) SA->pymax = SA->pxmax;
    }
  } else {
    FILE2_open (fp, PP->input_fname, "r", EXIT);
    while ( 1 ){  // compute xmax, ymax
      x = (SEQ_LONG)FILE2_read_int (&fp);
      y = (SEQ_LONG)FILE2_read_int (&fp);
      FILE2_read_until_newline (&fp);
      if ( FILE_err&2 ) break;
      ENMAX (SA->xmax, x);
      ENMAX (SA->ymax, y);
    }
  }
  FILE2_close (&fp);

  SS->str_len_org = SA->col_num==1? SA->xmax: (SA->col_num==-1?MAX(SA->xmax,SA->ymax):SA->ymax);
  if ( SA->pymax == -1 ){
    SA->plen = SA->pxmax;  // length of 1 dot
    if ( SA->line_skip > 0 ){
      SA->pxmax = SA->pymax = (SEQ_LONG)ceil(sqrt(SS->str_len_org*SA->line_skip/SA->plen));
      SS->str_len = (SS->str_len_org+SA->plen-1) / SA->plen;  // total length of line
    } else {
      SA->pxmax = (SA->xmax+SA->plen-1)/SA->plen;
      SA->pymax = (SA->ymax+SA->plen-1)/SA->plen;
	  	SA->boxx = (SA->xmax+SA->pxmax)/SA->pxmax;
      SA->boxy = (SA->ymax+SA->pymax)/SA->pymax;
    }
  } else {
    if ( SA->pymax == 0 && SA->xmax>0 ){
        SA->pymax = (SEQ_LONG)((((double)SA->pxmax)*SA->ymax)/SA->xmax);
    }
    if ( SA->xmax > 0 && SA->pxmax>0 && SA->pymax>0 ){
	  	SA->boxx = (SA->xmax+SA->pxmax)/SA->pxmax;
      SA->boxy = (SA->ymax+SA->pymax)/SA->pymax;
    }
    if ( SA->line_skip>0 ){
      SS->str_len = SA->pxmax*((SA->pymax+SA->line_skip-1)/SA->line_skip);
      SA->plen = SS->str_len_org/SS->str_len;
    }
  }
  if ( SA->line_skip > 0 ) SA->plen = SS->str_len_org/SS->str_len;  // length of 1 dot
}

/* draw grid line to cell */
void SACHICAm_draw_grid (SACHICA *SA){
  FILE2 fp;
  int x, z;
  
  FILE2_open (fp, SA->PP.position_fname, "r", EXIT);
  z = (int)FILE2_read_int (&fp);
  FILE2_read_until_newline (&fp);
  while (1){
    z = (int)FILE2_read_int (&fp);
    FILE2_read_until_newline (&fp);
    if ( z/SA->boxy>=SA->pymax || z/SA->boxx>= SA->pxmax ) break;
    if ( FILE_err&2 ) break;
    FLOOP (x, 0, SA->pxmax) SA->cell[x][z/SA->boxy] = SACHICA_CELL_END-1;
    ARY_FILL (SA->cell[z/SA->boxx], 0, SA->pymax, SACHICA_CELL_END-1);
  }
  FILE2_close (&fp);
}

 
// increment the cell corresponding to (x,y)
void SACHICAm_cell_inc (SACHICA *SA, SEQ_LONG x, SEQ_LONG y){
  int xx, yy;
  double z;
  SEQ *SS = &SA->PP.SS;

  if ( x<0 || y< 0 ) return;
  if ( SA->line_skip > 0 ){
    if ( SA->col_num == -1 ){
      yy = (int)(((double)x)*SS->str_len / SS->str_len_org);
      xx = yy%SA->pxmax;
      yy = ((int)(yy/SA->pxmax)) *SA->line_skip;
      if ( SA->pxmax>xx && SA->pymax>yy && SA->cell[xx][yy]<SACHICA_CELL_END-1) SA->cell[xx][yy]++;
    }
    z = ((double)(SA->col_num==1? x: y)) *SS->str_len / SS->str_len_org;
    yy = (int)((double)(z / SA->pxmax)) * SA->line_skip;
    xx = (int)(z - (((int)(z/SA->pxmax))*SA->pxmax));
  } else { xx = x/SA->boxx; yy = y/SA->boxy; }
  if ( SA->pxmax>xx && SA->pymax>yy && SA->cell[xx][yy]<SACHICA_CELL_END-1) SA->cell[xx][yy]++;
}

// negative image
void SACHICAm_negative (SACHICA *SA){
  int x, y;
  FLOOP (x, 0, SA->pxmax){
    FLOOP (y, 0, SA->pymax)
        SA->cell[x][y] = SACHICA_CELL_END -1 - SA->cell[x][y];
  }
}

// clear the cells with small values
void SACHICAm_clear_small_value (SACHICA *SA){
  int x, y;
  FLOOP (y, 0, SA->pymax){
    FLOOP (x, 0, SA->pxmax)
        if ( SA->cell[x][y] < SA->th ) SA->cell[x][y] = 0;
  }
}

// clear the cells with small values
void SACHICAm_mirroring (SACHICA *SA){
  int x, y;
  FLOOP (y, 0, SA->pymax){
    FLOOP (x, 0, SA->pxmax)
        if ( y<SA->pxmax && x<SA->pymax ) ENMAX (SA->cell[x][y], SA->cell[y][x]);
  }
}

void SACHICAm_read_file_iter (SACHICA *SA, char *fname, double *den, int rflag){
  SEQ_LONG x, y, n=0, nn=0;
  int i, xx, yy=0;
  FILE2 fp;
  PROBLEM *PP = &SA->PP;
  SEQ *SS = &PP->SS;

  FILE2_open (fp, fname, "r", {free2(den);EXIT;});
  while ( 1 ){
    x = (SEQ_LONG)FILE2_read_int (&fp);
    y = (SEQ_LONG)FILE2_read_int (&fp);
    if ( FILE_err&2 ) break;
    FILE2_read_until_newline (&fp);
    PP->II.outputs++;
    if ( x<0 || y<0 ) continue;

    if ( PP->problem & SACHICAM_POSITION_SC ){
      xx = (int)(((double)x)*SS->str_len/SA->ymax);
      if ( xx == n ) ENMAX (yy, y);
      else { den[n] = yy; n = xx; yy = y; }
    } else if ( PP->problem & SACHICAM_READ_INTERVAL ){
      xx = (int)(((double)x)*SS->str_len/SA->ymax);
      yy = (int)(((double)y)*SS->str_len/SA->ymax);
      nn = (SEQ_LONG)(((double)xx)*SA->ymax / SS->str_len);
      FLOOP (i, xx, yy+1){
        if ( i>= (int)(SS->str_len/SA->ymax) ) break;
        n = (SEQ_LONG)((((double)i+1)*SA->ymax / SS->str_len));
        if ( n == nn ) den[i] = i>0? den[i-1]: 0;
        else den[i] += ((double)(MIN (n, y+1 ) - x)) / (n-nn);
        x = nn = n;
      }
    } else {
      if ( rflag ){
        if ( PP->problem & SACHICA_REVERSE_X ) x = SA->xmax - x;
        else y = SA->ymax - y;
      }
      SACHICAm_cell_inc (SA, x, y);
        // for the swap of x and y
      if ( SA->col_num==-1 && SA->line_skip==0 ) SACHICAm_cell_inc (SA, y, x);
    }

  }
  FILE2_close (&fp);
  if ( SA->th>0 ) SACHICAm_clear_small_value (SA);  // clear pixels of small intensity
  if ( PP->problem & SACHICAM_POSITION_SC ) den[n] = yy;  // for the last cell
}

/*******************************************************************************/
void SACHICAm_read_file (SACHICA *SA){
  double *den=NULL;
  int i, xx, yy;
  PROBLEM *PP = &SA->PP;
  SEQ *SS = &PP->SS;
  
  if ( SA->line_skip > 0 || (PP->problem & SACHICAM_READ_INTERVAL) ){
    malloc2 (den, SS->str_len, EXIT);
    ARY_FILL (den, 0, SS->str_len, 0);
  }

  SACHICAm_read_file_iter (SA, PP->input_fname, den, 0);
  if (PP->problem & (SACHICAM_READ_INTERVAL + SACHICAM_POSITION_SC) ){
    for (i=yy=0 ; yy<SA->pymax ; yy+=SA->line_skip){
      FLOOP (xx, 0, SA->pxmax){
        SA->cell[xx][yy] = (SACHICA_CELL_CHAR)MIN (255, den[i] * SA->density);
        i++;
        if ( PP->problem & SACHICAM_NEGATIVE ) SA->cell[xx][yy] = 255-SA->cell[xx][yy];
      }
    }
    free2 (den);
  }

  if ( !PP->position2_fname || SA->line_skip>0 ) return;
  SACHICAm_read_file_iter (SA, PP->position2_fname, den, 1);
}

// write a pgm/bmp file
void SACHICAm_output (SACHICA *SA, int bmp){
  PROBLEM *PP = &SA->PP;
  int x, y, yy, m=0, z, f= (PP->problem & SACHICAM_NEGATIVE) && SA->line_skip == 0;
  unsigned char *buf;
  BMPFILEHEADER fileHeader;
  BMPINFOHEADER infoHeader;
  FILE *fp;
  
  if ( SA->th == 0 ){ // adjasting intensity 
    FLOOP (y, 0, SA->pymax){
      FLOOP (x, 0, SA->pxmax) ENMAX (m, SA->cell[x][y]);
    }
    if ( m == 0 ) return;
    SA->density *= ((double)256)/m;
  }

  fopen2 (fp, PP->output_fname, "wb", EXIT);
  malloc2 (buf, SA->pxmax*4+10, EXIT);

  if ( bmp ){
    fileHeader.bfType='M'*256+'B';  // means 'BM'
    fileHeader.bfSize=sizeof(BMPFILEHEADER)+sizeof(BMPINFOHEADER)+4*SA->pxmax*SA->pymax;
    fileHeader.bfReserved1=0;
    fileHeader.bfReserved2=0;
    fileHeader.bfOffBits=sizeof(BMPFILEHEADER)+sizeof(BMPINFOHEADER);
    fwrite (&fileHeader,1,sizeof(BMPFILEHEADER), fp);
  
    infoHeader.biSize=sizeof(BMPINFOHEADER);
    infoHeader.biWidth=SA->pxmax;
    infoHeader.biHeight=SA->pymax;
    infoHeader.biPlanes=1;
    infoHeader.biBitCount=32;
    infoHeader.biCompression=0;
    infoHeader.biSizeImage=4*SA->pxmax*SA->pymax;
    infoHeader.biXPixPerMeter=3780; // 96dpi
    infoHeader.biYPixPerMeter=3780;
    infoHeader.biClrUsed=0;
    infoHeader.biClrImportant=0;

    fwrite (&infoHeader,1,sizeof(BMPINFOHEADER), fp);
  } else {
    fprintf (fp, "P5\n");
    fprintf (fp, "#result of comparison\n");
    fprintf (fp, "%d %d\n", SA->pxmax, SA->pymax);
    fprintf (fp, "255\n");
  }

  FLOOP (yy, 0, SA->pymax){
    y = bmp? SA->pymax-1-yy: yy;
    FLOOP (x, 0, SA->pxmax){
      if ( SA->th > 0 ){ // for threshold drawing 
        if ( SA->cell[x][y] >= SA->th ){
          z = 255;
          PP->II.outputs2++;
        } else z = 0;
      } else {  // intensity drawing
        z = (unsigned char)MIN (255, SA->density*SA->cell[x][y]);
      }
      if ( f ) z = 255-z;
      if ( bmp ){ buf[x*4] = buf[x*4+1] = buf[x*4+2] = z; buf[x*4+3] = 0; }
      else buf[x] = z;
    }
    fwrite (buf, bmp?4:1, SA->pxmax, fp);   // write 1 line to the file
  }
  free (buf);
  fclose (fp);
}

// write a pgm file
void SACHICAm_emphasize (SACHICA *SA, int k){
  int x, y, xx, yy;
  FLOOP (y, 0, SA->pymax-k){
    FLOOP (x, 0, SA->pxmax-k){
      FLOOP (yy, 0, k+1){
        FLOOP (xx, 0, k+1) ENMAX (SA->cell[x][y], SA->cell[x+xx][y+yy]);
      }
    }
  }
}

/****************************************/
/*  detect homologinous intervals */
/****************************************/
void SACHICAm_interval_detect (SACHICA *SA){
  int x, y, i=0, h=0, xx, yy, flag=0, start=1, end=0;
  SEQ *SS = &SA->PP.SS;

  for (y=0 ; y<SA->pymax ; y+=SA->line_skip){
    FLOOP (x, 0, SA->pxmax){
      if ( SA->cell[x][y] < SA->th ) h++;
      if ( i >= SS->plen ){
        if ( x < SS->plen ){ xx = x +(SA->pxmax -SS->plen) ; yy = y -SA->line_skip; }
        else { xx = x-SS->plen; yy = y; }
        if ( SA->cell[xx][yy] < SA->th ) h--;
      }
      if ( i >= SS->plen-1 ){
        if ( h <= SS->perr ){
          if ( flag==0 ){
            flag = 1;
            if ( end < i -SS->plen && end-start >= SA->th2 ){
              while ( SA->cell[start%SA->pxmax][(start/SA->pxmax)*SA->line_skip] < SA->th ) start++;
              while ( end>0 && SA->cell[(end-1)%SA->pxmax][((end-1)/SA->pxmax)*SA->line_skip] < SA->th ) end--;
              fprintf (SA->PP.II.fp, SEQ_LONGF " " SEQ_LONGF " : %d\n",
                  (SEQ_LONG)(((double)start) /SS->str_len *SS->str_len_org),
                  (SEQ_LONG)(((double)end) /SS->str_len *SS->str_len_org),
                  end-start);
              start = i;
            }
          }
        } else if ( flag == 1 ){ end = i; flag = 0; }
      }
      i++;
    }
  }
}

/****************************************/
/*  detect homologinous intervals */
/* solve a DP to find maximal numbers of dots sequencially trace-able with jumping to 3 cells for each skip */
/* if #connected dots > SACHICA_th, output */
/****************************************/
void SACHICAm_detect (SACHICA *SA){
  PROBLEM *PP = &SA->PP;
  SEQ *SS = &PP->SS;
  short **cel, **celx=0, **cely=0;
  int cnt=0, ccc1, ccc2;
  int i, x, y, xx, yy;
  int dirmax = (PP->problem&SACHICAM_LONG_SKIP)?31:10;
  int dirx1[10] = {1,0, 2,1,0, 2,1, 3,2,1};
  int diry1[10] = {0,1, 0,1,2, 1,2, 1,2,3};
  int dirx2[31] = {1,0, 2,1,0, 2,1, 3,2,1, 4,3,2,1, 5,4,3,2,1, 5,4,3,2, 6,5,4,3,2, 5,4, 5 };
  int diry2[31] = {0,1, 0,1,2, 1,2, 1,2,3, 1,2,3,4, 1,2,3,4,5, 2,3,4,5, 2,3,4,5,6, 4,5, 5 };
  int *dirx=dirx1, *diry=diry1;

  if ( PP->problem & SACHICAM_LONG_SKIP ){ dirx = dirx2; diry = diry2; }

  calloc2d (cel, SA->pxmax+10, SA->pymax+10, goto END2);
  malloc2d (celx, SA->pxmax+10, SA->pymax+10, goto END2);
  malloc2d (cely, SA->pxmax+10, SA->pymax+10, goto END2);
  FLOOP (x, 0, SA->pxmax+10){
    ARY_FILL (celx[x], 0, SA->pymax+10, -1);
    ARY_FILL (cely[x], 0, SA->pymax+10, -1);
  }
  BLOOP (x, SA->pxmax, 0){  
    BLOOP (y, SA->pymax, 0){
      if ( SA->cell[x][y] > 0 ){
        cel[x][y] = 1;
        celx[x][y] = x;
        cely[x][y] = y;
        FLOOP (i, 0, dirmax){
          if ( cel[x+dirx[i]][y+diry[i]]+1 > cel[x][y] ){
            cel[x][y] = cel[x+dirx[i]][y+diry[i]]+1;  // cel[x][y] := longest dot-sequence to right-down direction from (x,y)
            celx[x][y] = celx[x+dirx[i]][y+diry[i]];  // x-coordinate of the end of the sequence
            cely[x][y] = cely[x+dirx[i]][y+diry[i]];  // y-coordinate of the end of the sequence
          }
        }
      }
    }
  }

  FLOOP (x, 0, SA->pxmax){
    FLOOP (y, 0, SA->pymax){
      if ( cel[x][y]> 0 ){
        FLOOP (i, 0, dirmax){
          if ( cel[x+dirx[i]][y+diry[i]] > 0 && 
               cel[x+dirx[i]][y+diry[i]] < cel[x][y] )
                  cel[x+dirx[i]][y+diry[i]] = cel[x][y]; // cel[x][y] := longest dot-sequence to left-up to right-down direction passing through (x,y)
        }
      }
    }
  }
  
/*
  for (x=SA->pxmax-1 ; x>=0 ; x-- ){
    for (y=SA->pymax-1 ; y>=0 ; y-- ){
      if ( cel[x][y] > 0 ){
        if ( cel[celx[x][y]][cely[x][y]] > cel[x][y] ){
          if ( cel[x][y] < SACHICA_th ) SACHICA_mdz[x][y] = 0;
          cel[x][y] = 0;
        }
      }
    }
  }
*/

  BLOOP (x, SA->pxmax, 0){
    BLOOP (y, SA->pymax, 0){
      if ( cel[x][y] > 0 ){
        FLOOP (i, 0, dirmax)
          if ( cel[x+dirx[i]][y+diry[i]] >0 ) 
             cel[x+dirx[i]][y+diry[i]] *= -1;
      }
    }
  }

  BLOOP (x, SA->pxmax, 0){
    BLOOP (y, SA->pymax, 0){
      xx = celx[x][y];
      yy = cely[x][y];
      if ( xx<0 || yy<0 ) continue;
      if ( cel[x][y] >= SA->th && cel[x][y] == -cel[xx][yy] &&
         xx-x >= SA->th2 && yy-y >= SA->th2 ){
       if ( SA->th2>0 && ( (xx-x)/(yy-y) >5 || (yy-y)/(xx-x)>5) ) goto END;

// printf ("%d %d   (%d) %d %d\n", x, y, cel[x][y], celx[x][y], cely[x][y]);
        cel[xx][yy] *= -1;
        cnt++;
        
        // checking closed to boundary or not (without -B option, ratio is 1.0 so any is accepted
        ccc1=0; ccc2=0;
        if ( ((double)x)/((double)SA->pxmax) < SS->err_ratio ) ccc1=1;
        if ( ((double)y)/((double)SA->pymax) < SS->err_ratio ) ccc1=1;
        if ( ((double)(SA->pxmax-1-xx))/((double)SA->pxmax) < SS->err_ratio ) ccc2=1;
        if ( ((double)(SA->pymax-1-yy))/((double)SA->pymax) < SS->err_ratio ) ccc2=1;
        if ( ccc1+ccc2>=SA->th1 ){   // #boundaries touching is no less than threshold
          PP->II.outputs2++;
          if ( PP->problem & SACHICAM_DETECT ){
            fprintf (PP->II.fp, SEQ_LONGF "," SEQ_LONGF " " SEQ_LONGF "," SEQ_LONGF "   %d\n",
                  x*SA->boxx, y*SA->boxy, (xx+1)*SA->boxx, (yy+1)*SA->boxy, (int)cel[x][y]);
          } else if ( PP->problem & SACHICAM_HOMOLOGY_REVERSE_X ){
            fprintf (PP->II.fp, SEQ_LONGF "," SEQ_LONGF " " SEQ_LONGF "," SEQ_LONGF "   %d\n",
                  SA->xmax -x*SA->boxx, y*SA->boxy, SA->xmax -(xx+1)*SA->boxx, (yy+1)*SA->boxy,
                  (int)cel[x][y]);
          } else if ( PP->problem & SACHICAM_HOMOLOGY_REVERSE_Y ){
            fprintf (PP->II.fp, SEQ_LONGF "," SEQ_LONGF " " SEQ_LONGF "," SEQ_LONGF "   %d\n",
                   x*SA->boxx, SA->ymax -y*SA->boxy, (xx+1)*SA->boxx, SA->ymax -(yy+1)*SA->boxy,
                   (int)cel[x][y]);
          }
        }
        END:;
      }
    }
  }

  FLOOP (x, 0, SA->pxmax){
    FLOOP (y, 0, SA->pymax)
        SA->cell[x][y] = abs(cel[x][y])<SA->th? 0: abs(cel[x][y]);
  }
  
  FLOOP (x, 0, SA->pxmax){
    FLOOP (y, 0, SA->pymax){
      if ( celx[x][y]<0 || cely[x][y]<0 ){ SA->cell[x][y]=0; continue; }
      xx = abs (x-celx[x][y]);
      yy = abs (y-cely[x][y]);
      if ( abs(cel[x][y]) >= SA->th ){
        if ( cel[x][y]<0 || 
         ( ( xx >= SA->th2 && yy >= SA->th2 ) &&
          ( SA->th2==0 || ( xx/yy<6 && yy/xx < 6) )) ){
          FLOOP (i, 0, dirmax)
            if ( cel[x+dirx[i]][y+diry[i]] >0 ) 
               cel[x+dirx[i]][y+diry[i]] *= -1;
          if ( cel[x][y] >0 ) cel[x][y] *= -1;
          SA->cell[x][y] = SACHICA_CELL_END-1;
        }
      }
    }   
  }

  FLOOP (x, 0, SA->pxmax){
    FLOOP (y, 0, SA->pymax)
        if ( cel[x][y] > 0 ) SA->cell[x][y] = 0;
  }

  END2:;
  if ( PP->II.outputs2 == 0 ) remove (PP->sc_fname);
  free2d (cel);
  free2d (celx);
  free2d (cely);
}

/***********************************************************************/
int SACHICAm_main (int argc, char *argv[]){
  SACHICA SA;
  PROBLEM *PP = &SA.PP;

  PROBLEM_init (PP);
  SACHICAm_read_param (&SA, argc, argv);  
  PROBLEM_alloc (PP, 0, 0, 0, 0, 0);
  if ( ERROR_MES ) goto END;
  SACHICAm_scan_file (&SA, PP->input_fname);
  if ( SA.xmax == 0 ) goto END;

  calloc2d (SA.cell, SA.pxmax+10, SA.pymax+10, EXIT);
  print_err ("sachicam:x:"SEQ_LONGF ", y:"SEQ_LONGF", px:%d, py:%d, xcell:%d, ycell:%d\n",
     SA.xmax, SA.ymax, SA.pxmax, SA.pymax, SA.boxx, SA.boxy);
  SACHICAm_read_file (&SA);
  
  if ( PP->sc_fname ){
    fopen2 (SA.PP.II.fp, PP->sc_fname, "w", EXIT);
    if ( SA.line_skip > 0 ) SACHICAm_interval_detect (&SA);
    else SACHICAm_detect (&SA);
    fclose (SA.PP.II.fp);
    if ( PP->II.outputs2 == 0 ) goto END;  // not draw picture if no homology string is found
  }
  if ( PP->problem & SACHICAM_DRAW_GRID ) SACHICAm_draw_grid (&SA);

  if ( SA.emphasize>0 ) SACHICAm_emphasize (&SA, SA.emphasize);
  if ( PP->problem & SACHICAM_MIRROR ) SACHICAm_mirroring (&SA);

  SACHICAm_output (&SA, PP->problem & SACHICAM_BMP );

  print_err ("#pairs "LONGF", #dots "LONGF"\n", PP->II.outputs, PP->II.outputs2);
  
  END:;
  free2d (SA.cell);
  PROBLEM_end (PP);
  return (0);
}

/*******************************************************************************/
#ifndef _NO_MAIN_
#define _NO_MAIN_
int main (int argc, char *argv[]){
  return (SACHICAm_main (argc, argv) );
}
#endif
/*******************************************************************************/



#endif



